@@ -2,8 +2,11 @@
 # Copyright 2018 Open Source Geospatial Foundation and others    #
 # licensed under MIT, Please consult LICENSE.txt for details     #
 ##################################################################
+import json
 from pathlib import PurePath
 
+from pywps.inout.formats import Supported_Formats
+from pywps.inout.types import Translations
 from pywps.translations import lower_case_dict
 from io import StringIO
 import os
@@ -203,11 +206,12 @@ class IOHandler(object):
         """
 
         validate = self.validator
-        _valid = validate(self, self.valid_mode)
-        if not _valid:
-            self.data_set = False
-            raise InvalidParameterValue('Input data not valid using '
-                                        'mode {}'.format(self.valid_mode))
+        if validate is not None:
+            _valid = validate(self, self.valid_mode)
+            if not _valid:
+                self.data_set = False
+                raise InvalidParameterValue('Input data not valid using '
+                                            'mode {}'.format(self.valid_mode))
         self.data_set = True
 
     @property
@@ -316,6 +320,14 @@ class IOHandler(object):
         WARNING: may be bytes or str"""
         return self._iohandler.data
 
+    def data_as_json(self):
+        # applies json.loads if needed
+        data = self._iohandler.data
+        if data and not isinstance(self._iohandler, DataHandler) and self.extension in ['.geojson', 'json']:
+            data = json.loads(data)
+            self.data = data  # switch to a DataHandler
+        return data
+
     @data.setter
     def data(self, value):
         self._iohandler = DataHandler(value, self)
@@ -460,7 +472,10 @@ class DataHandler(FileHandler):
             openmode = self._openmode(self.data)
             kwargs = {} if 'b' in openmode else {'encoding': 'utf8'}
             with open(self._file, openmode, **kwargs) as fh:
-                fh.write(self.data)
+                if isinstance(self.data, (bytes, str)):
+                    fh.write(self.data)
+                else:
+                    json.dump(self.data, fh)
 
         return self._file
 
@@ -643,8 +658,8 @@ class BasicIO:
         self.title = title
         self.abstract = abstract
         self.keywords = keywords
-        self.min_occurs = int(min_occurs)
-        self.max_occurs = int(max_occurs)
+        self.min_occurs = int(min_occurs) if min_occurs is not None else 0
+        self.max_occurs = int(max_occurs) if max_occurs is not None else None
         self.metadata = metadata
         self.translations = lower_case_dict(translations)
 
@@ -715,7 +730,7 @@ class BasicComplex(object):
     def validator(self):
         """Return the proper validator for given data_format
         """
-        return self.data_format.validate
+        return None if self.data_format is None else self.data_format.validate
 
     @property
     def supported_formats(self):
@@ -1071,8 +1086,8 @@ class ComplexOutput(BasicIO, BasicComplex, IOHandler):
     """
 
     def __init__(self, identifier, title=None, abstract=None, keywords=None,
-                 workdir=None, data_format=None, supported_formats=None,
-                 mode=MODE.NONE, translations=None):
+                 workdir=None, data_format=None, supported_formats: Supported_Formats = None,
+                 mode=MODE.NONE, translations: Translations = None):
         BasicIO.__init__(self, identifier, title, abstract, keywords, translations=translations)
         IOHandler.__init__(self, workdir=workdir, mode=mode)
         BasicComplex.__init__(self, data_format, supported_formats)
